#############################################################################
##
#A  mop.g            MOP                      Ulf Juergens, Gerhard Roehrle
##
#A  $Id: mop.g,v 1.16 2001/11/25 15:51:13 ulf Exp $
##
#Y  Copyright (C) 2000, 2001  Ulf Juergens, Gerhard Roehrle.
##
##  MOP computes upper bounds for the modality of the action of
##  parabolic subgroups P of algebraic groups with simply laced Dynkin diagram
##  on the Lie algebra of the unipotent radical of P.
##


#############################################################################
##
#F MOP.JHeight(<r>, <J>)
##
## returns the sum of the coefficients of the
## root <r> whose indices are not in the list <J>.
##
MOP.JHeight:=function(r, J)
    return Sum(Filtered([1..MOP.W.rank], i-> not i in J), j->r[j]);
end;


#############################################################################
##
#F MOP.restrictedOpr(<system>, <g>) . . . . . . . . . . . . . . . . . . .
##
## returns the images of a list <system> under an element <g> of the
## big Weyl group MOP.W, if these images are all positive, and otherwise the
## original list.
##
MOP.restrictedOpr:=function(system, g)
    local image, image2;

    image:=List(system, p->p^g);
    if IsSubset(MOP.RootNumbers, Set(image)) then
        return image;
    else
        return system;
    fi;
end;


#############################################################################
##
#F MOP.PrintSubsystem(<sub>) . . . . . . . . . . . . . . . . . . . . .
##
## displays a subsystem <sub> in the diagram record.
##
MOP.PrintSubsystem:=function(sub)
  local i;

  Print(sub.type[1].type,sub.type[1].rank);
  for i in [2..Length(sub.type)] do
    Print("x", sub.type[i].type, sub.type[i].rank);
  od;
  Print(" J=", sub.J);
  if IsBound(sub.b) then
    Print(" inclusion=", sub.b);
  fi;
  Print("\n");
end;


#############################################################################
##
#F MOP.PrintType(<r>) . . . . . . . . . . . . . . . . . . . . . . . .
##
## is used for the diagram record.
##
MOP.PrintType:=function(r)
  Print(r.type, r.rank);
end;


#############################################################################
##
#F MOP.SubSystems:=function(<type>, <rank>)
##
## returns a list of subsystems of the root system of type <type><rank>
## that are then used in -->MOP.ComputeInductionList.
##
MOP.SubSystems:=function(type, rank)
  if type="A" then
    if rank=1 then
      return [];
    else
      return [rec(type:=[rec(type:="A", rank:=rank-1)],
              incl:=IdentityMat(rank){[1..rank-1]})];
    fi;
  elif type="D" then
    if rank=1 then
      return [];
    else
      return [rec(type:=[rec(type:="A", rank:=rank-1)],
              incl:=IdentityMat(rank){Concatenation([1],[3..rank])}),
              rec(type:=[rec(type:="A", rank:=rank-1)],
              incl:=IdentityMat(rank){[2..rank]}),
              rec(type:=[rec(type:="D", rank:=rank-1)],
              incl:=IdentityMat(rank){[1..rank-1]})];
    fi;
  elif type="E" then
    if rank=6 then
      return [
        rec(
          type:=[rec(type:="D",  rank:=5)],
          incl:=[[0,0,0,0,1,0],
                 [0,1,0,0,0,0],
                 [0,0,0,1,0,0],
                 [0,0,1,0,0,0],
                 [1,0,0,0,0,0]]),
        rec(type:=[rec(type:="A", rank:=1),
                   rec(type:="A", rank:=5)],
          incl:=[[1,2,2,3,2,1],
                 [1,0,0,0,0,0],
                 [0,0,1,0,0,0],
                 [0,0,0,1,0,0],
                 [0,0,0,0,1,0],
                 [0,0,0,0,0,1]])];
    elif rank=7 then
      return [
        rec(type:=[rec(type:="E", rank:=6)],
          incl:=[[1,0,0,0,0,0,0],
                 [0,1,0,0,0,0,0],
                 [0,0,1,0,0,0,0],
                 [0,0,0,1,0,0,0],
                 [0,0,0,0,1,0,0],
                 [0,0,0,0,0,1,0]]),
        rec(type:=[rec(type:="A", rank:=7)],
          incl:=[[1,0,0,0,0,0,0],
                 [0,0,1,0,0,0,0],
                 [0,0,0,1,0,0,0],
                 [0,0,0,0,1,0,0],
                 [0,0,0,0,0,1,0],
                 [0,0,0,0,0,0,1],
                 [1,2,2,3,2,1,0]]),
      rec(type:=[rec(type:="A", rank:=1),
                 rec(type:="D", rank:=6)],
        incl:=[[2,2,3,4,3,2,1],
               [0,0,1,0,0,0,0],
               [0,1,0,0,0,0,0],
               [0,0,0,1,0,0,0],
               [0,0,0,0,1,0,0],
               [0,0,0,0,0,1,0],
               [0,0,0,0,0,0,1]]),
      rec(type:=[rec(type:="A", rank:=2),
                 rec(type:="A", rank:=5)],
        incl:=[[1,2,3,4,3,2,1],
               [1,0,0,0,0,0,0],
               [0,1,0,0,0,0,0],
               [0,0,0,1,0,0,0],
               [0,0,0,0,1,0,0],
               [0,0,0,0,0,1,0],
               [0,0,0,0,0,0,1]])
    ];
  elif rank=8 then
    return [
      rec(type:=[rec(type:="A", rank:=8)],
          incl:=[[1,3,3,5,4,3,2,1],
                 [1,0,0,0,0,0,0,0],
                 [0,0,1,0,0,0,0,0],
                 [0,0,0,1,0,0,0,0],
                 [0,0,0,0,1,0,0,0],
                 [0,0,0,0,0,1,0,0],
                 [0,0,0,0,0,0,1,0],
                 [0,0,0,0,0,0,0,1]]),
      rec(type:=[rec(type:="D", rank:=8)],
          incl:=[[0,0,1,0,0,0,0,0],
                 [0,1,0,0,0,0,0,0],
                 [0,0,0,1,0,0,0,0],
                 [0,0,0,0,1,0,0,0],
                 [0,0,0,0,0,1,0,0],
                 [0,0,0,0,0,0,1,0],
                 [0,0,0,0,0,0,0,1],
                 [2,2,3,4,3,2,1,0]]),
      rec(type:=[rec(type:="A", rank:=1),
                 rec(type:="E", rank:=7)],
          incl:=[[2,3,4,6,5,4,3,2],
                 [1,0,0,0,0,0,0,0],
                 [0,1,0,0,0,0,0,0],
                 [0,0,1,0,0,0,0,0],
                 [0,0,0,1,0,0,0,0],
                 [0,0,0,0,1,0,0,0],
                 [0,0,0,0,0,1,0,0],
                 [0,0,0,0,0,0,1,0]]),
      rec(type:=[rec(type:="A", rank:=2),
                 rec(type:="E", rank:=6)],
          incl:=[[2,3,4,6,5,4,3,1],
                 [0,0,0,0,0,0,0,1],
                 [1,0,0,0,0,0,0,0],
                 [0,1,0,0,0,0,0,0],
                 [0,0,1,0,0,0,0,0],
                 [0,0,0,1,0,0,0,0],
                 [0,0,0,0,1,0,0,0],
                 [0,0,0,0,0,1,0,0]])
#      ,
#      rec(type:=[rec(type:="A", rank:=4),
#                 rec(type:="A", rank:=4)],
#          incl:=[[2,3,4,6,5,3,2,1],
#                   [0,0,0,0,0,1,0,0],
#                   [0,0,0,0,0,0,1,0],
#                   [0,0,0,0,0,0,0,1],
#                   [1,0,0,0,0,0,0,0],
#                   [0,0,1,0,0,0,0,0],
#                   [0,0,0,1,0,0,0,0],
#                   [0,1,0,0,0,0,0,0]])
      ];
    fi;
  fi;
end;


#############################################################################
##
#F MOP.ComputeInductionList() . . . . . . . . . . . . . . . .
##
## determines all conjugates of subsystems recruited from a file
## named <type><rank>.sub under the action of the big Weyl group MOP.W
## subject to the condition that the images of the positive roots of
## the subsystem are positive in the whole system.
##
## The parameters <type> and <rank> are taken from the MOP record.
##
MOP.ComputeInductionList:=function()
    local s, w, i, inductionlist, orbit, b, x, bJ;

    if not MOP.UseInductionList then
        MOP.InductionList:=[];
        return;
    fi;

    MOP.subsystems:=MOP.SubSystems(MOP.Type, MOP.W.rank);

    MOP.VerbosePrint(1, "computing InductionList\n");
    MOP.InductionList:=[];

    for s in MOP.subsystems do
        for i in s.type do
            i.operations:=MOP.TypeOps;
        od;
        w:=CoxeterGroup(ApplyFunc(DirectSumMat, 
                                  List(s.type, i->CartanMat(i.type, i.rank)))
                       );
        w.SimpleRootImages:=List(s.incl, i->Position(MOP.W.roots, i));
        orbit:=Orbit(MOP.W, w.SimpleRootImages, MOP.restrictedOpr);

        MOP.VerbosePrint(1, "number of subsystems of type ", s.type,
              ": ", Length(orbit), "\n");

        for b in orbit do
            bJ:=Filtered([1..Length(s.incl)],
                        i->b[i] in MOP.PositionsRootsInLevi);
            AddSet(MOP.SubDiagrams, rec(
                    operations:=MOP.SubsystemOps,
                    type:=s.type, J:=bJ));
            AddSet(MOP.SubDiagramsDetail, rec(
                    operations:=MOP.SubsystemOps,
                    b:=b, type:=s.type, J:=bJ));
        od;

        inductionlist:=List(orbit, v->
                List(w.roots{[1..w.N]}, r->
                     Position(MOP.W.roots, Sum([1..Length(r)], i->
                             r[i]*MOP.W.roots[v[i]]))));
        for x in inductionlist do
            Sort(x);
        od;
        inductionlist:=Set(inductionlist);
        MOP.VerbosePrint(1, "number of elements: ", Size(inductionlist), "\n");
        Append(MOP.InductionList, inductionlist);
    od;
end;


#############################################################################
##
#F MOP.OperationOnString(<s>, <g>) . . . . . . . . . . . . . . .
##
## permutes a string <s> under the action of an element <g> of the
## "small" Weyl group MOP.WJ.
##
MOP.OperationOnString:=function(s, r)
    local q;

    q:=BlistList(MOP.RootNumbers, []);
    return rec(operations:=MOP.StringOps,
               m:=Permuted(Concatenation(s.m,q), r){[1..MOP.W.N]},
               k:=Permuted(Concatenation(s.k,q), r){[1..MOP.W.N]});
end;


#############################################################################
##
#F MOP.SupportLength(<s>) . . . . . . . . . . . . . . . . . . . . .
##
## returns the length of the support of a string <s>.
##
MOP.SupportLength:=function(s)
    return SizeBlist(s.k)+SizeBlist(s.m);
end;


#############################################################################
##
#F MOP.kPositions(<s>) . . . . . . . . . . . . . . . . . . . . . . .
##
## returns the numbers of those roots in <s> marked with the letter "K".
##
MOP.kPositions:=function(s)
    return ListBlist([1..Length(s.k)], s.k);
end;


#############################################################################
##
#F MOP.mPositions(<s>) . . . . . . . . . . . . . . . . . . . . . . .
##
## returns the numbers of those roots in <s> marked with the letter "M".
##
MOP.mPositions:=function(s)
    return ListBlist([1..Length(s.m)], s.m);
end;


#############################################################################
##
#F MOP.PrintString(<s>) . . . . . . . . . . . . . . . . . . . . . .
##
## displays a string <s> by indicating the numbers of those roots in
## <s> marked with "M" or "K", respectively.
##
MOP.PrintString:=function(s)
  Print("M ", MOP.mPositions(s), "\n");
  Print("K ", MOP.kPositions(s), "\n");
end;


MOP.StringOps:=rec(Print:=MOP.PrintString,
                   operations:=rec(Print:=function(x) 
                                            Print("MOP.StringOps"); 
                                          end)
                  );


MOP.StringCopy:=function(s)
  return rec(m:=Copy(s.m), k:=Copy(s.k), operations:=MOP.StringOps);
end;

MOP.VerbosePrint:=function(arg)
  local x;

  if MOP.Verbose>=arg[1] then
    for x in arg{[2..Length(arg)]} do
      Print(x);
    od;
  fi;
end;


#############################################################################
##
#F MOP.StackSize() . . . . . . . . . . . . . . . . . . . . . . . .
##
## returns the number of elements in the stack.
##
MOP.StackSize:=function()
    return MOP.StackPointer-1;
end;


#############################################################################
##
#F MOP.StackEmpty() . . . . . . . . . . . . . . . . . . . . . . .
##
## returns <boolean> to indicate whether the stack is empty.
##
MOP.StackEmpty:=function()
    return MOP.StackSize()=0;
end;


#############################################################################
##
#F MOP.PushString(<s>, <stat>) . . . . . . . . . . . . . . . . . .
##
## inserts a string <s> into the stack. The second argument
## <stat> is a <boolean> value which is evaluated when a string is
## removed from the stack by MOP.PopString().
## If <stat> is <true>, then <s> will be analyzed, otherwise it is
## inserted in MOP.RedundancyList.
##
MOP.PushString:=function(s, stat)
    MOP.Stack[MOP.StackPointer]:=rec(str:=MOP.StringCopy(s), stat:=stat);
    MOP.StackPointer:=MOP.StackPointer+1;
end;


#############################################################################
##
#F MOP.PopString() . . . . . . . . . . . . . . . . . . . . . . . .
##
## removes a string from the stack and returns it, provided there
## is one, otherwise it returns <false>.
##
MOP.PopString:=function()
    if MOP.StackEmpty() then
        return false;
    else
        MOP.StackPointer:=MOP.StackPointer-1;
        return MOP.Stack[MOP.StackPointer];
    fi;
end;


#############################################################################
##
#F MOP.WriteStack() . . . . . . . . . . . . . . . . . . . . . . .
##
## saves the stack and along with it some more global variables to a file
## so that the results computed thus far are not lost in case of a system
## crash. When the program is started after recovery, it attempts to
## restore the saved status by means of MOP.ReadStack.
##
MOP.WriteStack:=function()
    local tmpfile, savefile;

    tmpfile:=Concatenation(MOP.JobName, ".tmp");
    savefile:=Concatenation(MOP.JobName, ".save");

    Unbind(MOP.StringOps.Print);

    PrintTo(tmpfile,
            "MOP.CannotAnalyzeList:=\n", MOP.CannotAnalyzeList, ";;\n",
            "MOP.factorlist:=\n", MOP.factorlist, ";;\n",
            "MOP.Counter:=", MOP.Counter, ";;\n",
            "MOP.MaxModality:=", MOP.MaxModality, ";;\n",
            "MOP.StackPointer:=", MOP.StackPointer, ";;\n",
            "MOP.RedundancyList:=Set(\n", MOP.RedundancyList, ");;\n",
            "MOP.Stack:=\n", MOP.Stack, ";;\n");

    MOP.StringOps.Print:=MOP.PrintString;

    Exec(Concatenation("mv ", tmpfile, " ", savefile));
end;


#############################################################################
##
#F MOP.ReadStack() . . . . . . . . . . . . . . . . . . . . . . . .
##
MOP.ReadStack:=function()
    return READ(Concatenation(MOP.JobName, ".save"));
end;


#############################################################################
##
#F MOP.ComputeRank(<s>) . . . . . . . . . . . . . . . . . . . . . .
##
## returns the number of linearly independent
## roots in the support of the string <s>.
##
MOP.ComputeRank:=function(s)
    local r;

    if MOP.SupportLength(s)=0 then
        return 0;
    else
        r:=RankMat(MOP.W.roots{ListBlist(MOP.RootNumbers,
                       UnionBlist(s.k, s.m))});
        return r;
    fi;
end;


#############################################################################
##
#F MOP.AlreadyDone(<s>, <normalmode>) . . . . . . . . . . . . . . . . .
##
## determines whether a string <s> is a "substring" (as a
## subvariety) of a string which has already been dealt with,
## i.e. which is taken from MOP.RedundancyList.
##
## For this and the following functions, the boolean parameter
## <normalmode> is passed. Only in the "normal" mode the counters
## are incremented and messages printed.
##
MOP.AlreadyDone:=function(s, normalmode)
    local i;

    for i in MOP.RedundancyList do
        if MOP.IsSubstring(i,s) then
            MOP.Counter.Redundancy:=MOP.Counter.Redundancy+1;
            if normalmode then
              MOP.VerbosePrint(1, "already done:\n",
                               i, "\n\n");
            fi;
            return true;
        fi;
    od;
    return false;
end;


#############################################################################
##
#F MOP.FixedJHeight(<s>, <normalmode>) . . . . . . . . . . . . . . . .
##
## determines whether the support of the string <s> consists
## entirely of roots of a fixed J-height h.
## [The number of L-orbits on the sum of root spaces relative to all roots
## of a fixed J-height is finite (thanks to a Theorem due to R.W.Richardson).]
##
MOP.FixedJHeight:=function(s, normalmode)
    local h;

    for h in [1..Length(MOP.JHeightStrings)] do
        if MOP.IsSubstring(MOP.JHeightStrings[h],s) then
          MOP.Counter.JHeight:=MOP.Counter.JHeight+1;
          if normalmode then
            MOP.VerbosePrint(1, "J-Height ", h, "\n");
          fi;
          return true;
        fi;
    od;
    return false;
end;


#############################################################################
##
#F MOP.MemberOfInductionList(<s>, <normalmode>) . . . . . . . . . . . . . . 
##
## determines whether a string <s> is contained in the induction list.
##
MOP.MemberOfInductionList:=function(s, normalmode)
    local i, support, l;

    if not MOP.UseInductionList then
      return false;
    fi;

    support:=ListBlist(MOP.RootNumbers, UnionBlist(s.k, s.m));
    l:=ListBlist([1..Length(MOP.InductionList)], 
                 IntersectionBlist(MOP.InductionListIndex{support}));

    if l=[] then
      return false;
    fi;

    MOP.Counter.Induction:=MOP.Counter.Induction+1;
    if normalmode then
      MOP.VerbosePrint(1, "induction list\n");
    fi;
    return true;
end;


#############################################################################
##
#F MOP.RankCondition(<s>, <normalmode>) . . . . . . . . . . . . . . . .
##
## determines whether enough nonzero positions of the
## given string <s> can be scaled to 1 using the action of the torus T
## with at most LowerBound nonzero positions
## remaining. Therefore, the modality of the action of P on the set of
## P-orbits passing through <s> is at most LowerBound.
##
## The first condition in the if-clause is less strict than the
## second one, but it ensures that the time-costly function MOP.ComputeRank
## is only invoked when there is a numerical chance to satisfy this
## criterion.
##
MOP.RankCondition:=function(s, normalmode)
    local mk;

    mk:=MOP.SupportLength(s);
    if mk<=MOP.W.rank+MOP.LowerBound and
       mk<=MOP.ComputeRank(s)+MOP.LowerBound then
          MOP.Counter.Rank:=MOP.Counter.Rank+1;
          if normalmode then
            MOP.VerbosePrint(1, "rank condition\n");
          fi;
          return true;
    else
      return false;
    fi;
end;


#############################################################################
##
#F MOP.IsSubstring(<s>, <t>) . . . . . . . . . . . . . . . . . . . . . . . . .
##
## decides whether a string <t> is contained in a string <s> as a substring.
##
                                                  #  s mk0mk0mk0
MOP.IsSubstring:=function(s, t)                   #  t mmmkkk000
    return                                        #    ---------
      IsSubsetBlist(s.k, t.k) and                 #    111010111
      IsSubsetBlist(t.m, s.m) and                 #    111011011
      IsSubsetBlist(UnionBlist(s.k, s.m), t.m);   #    110111111
end;                                              #    ---------
                                                  #    110010011

#############################################################################
##
#F MOP.ComputeOperation(<rs>) . . . . . . . . . . . . . . . . . . .
##
## returns the operation of the group <rs> on the root system as a record
##
##   rec (pairs, fromto) .
##
## The operation of a root subgroup relative to a root r is represented
## by a list of pairs of roots (b,c) where c=b+r.
## For each root r in the support of the parabolic subgroup P
## we create the list of all pairs on which the
## corresponding root subgroup acts as above. These lists are stored in
## operation.pairs, and each list is stored in operation.fromto[b][c]
## for each pair (b,c) in that list.
##
MOP.ComputeOperation:=function(rs)
    local op, i, l, ll, x, operation, opnumber;

    operation:=rec();
    operation.rootnumbers:=Concatenation([1..rs.N],
      List(-Filtered(rs.roots{[1..rs.N]}, r->MOP.JHeight(r, MOP.J)=0),
           r->Position(rs.roots, r)));
    operation.pairs:=[];
    operation.fromto:=List(MOP.RootNumbers, x->
                        List(MOP.RootNumbers, y->0));

    for opnumber in operation.rootnumbers do
        op:=rs.roots[opnumber];
        ll:=[];

        l:=List(MOP.RootNumbers, r->Position(rs.roots, rs.roots[r]+op));
        for i in MOP.RootNumbers do
            if l[i]<>false and
              rs.roots[i] in MOP.RootsInRadical then
                Add(ll, [i, l[i]]);
            fi;
        od;

        operation.pairs[opnumber]:=ll;
        for x in ll do
            operation.fromto[x[1]][x[2]]:=Position(rs.roots, op);
        od;
    od;
    return operation;
end;


#############################################################################
##
#F MOP.ReduceRow(<m>, <row>) . . . . . . . . . . . . . . . . . . . . .
##
## is part of the Gaussian elimination procedure for matrices
## with polynomial entries (--> MOP.Reduce).
## It searches for a nonzero entry in a given <row> of a matrix <m>.
## If there exists such an entry, it eliminates all entries below and
## returns the leading coefficient, otherwise it returns <false>.
##
MOP.ReduceRow:=function(m, row, simpleops, ColumnList)
  local NonzeroColumns, i, j, factor, column;

  NonzeroColumns:=Filtered(simpleops, col->
                    m[row][col] <> MOP.MultiZero and
                    IsMonomial(m[row][col]) and
                    Position(String(m[row][col]), 'k') = false);

  if NonzeroColumns=[] then
    return false;
  fi;

  column:=Minimum(NonzeroColumns);
  ColumnList[row]:=column;

  for i in [1..Length(m)] do
    if i<>row then
      factor:=m[i][column];
      if factor <> MOP.MultiZero then
        for j in [1..Length(m[i])] do
          m[i][j]:=m[i][j]*m[row][column]-m[row][j]*factor;
        od;
      fi;
    fi;
  od;
  return m[row][column].polynomial[1].Bscalar;
end;


#############################################################################
##
#F MOP.Reduce(<string>, <target>, <record>) . . . . . . . . . . . . . . . .
##
## tries to eliminate a <target> entry in a string s
## by means of certain operators.
##
## It returns a record with three fields:
##   eliminationlist is a list of positions which can be eliminated
##   pollutionlist   is a list of positions which are set to "K" by
##                   the operations
##   factorlist      is a list of primes which are not allowed as the
##                   characteristic of the underlying field in order to
##                   perform the elimination
##
MOP.Reduce:=function(s, target, record)
  local r, op, eliminationlist, pollutionlist, factorlist, factor, i, j,
        x, CoveredRows, p, nr, matrix, matrixop, simpleops, kpos,
        success, oldlist, newlist, ColumnList;

  ColumnList:=[];

##
## CoveredRows are all those roots in the string s that are
## labelled either by "0" or "M" and which are an image
## under any operation.
##
  matrix:=Copy(record.matrix);
  matrixop:=Copy(record.matrixoperators);
  simpleops:=Filtered([1..Length(matrixop)],
        	      j->Length(matrixop[j])=1);
  kpos:=[];
  CoveredRows:=Set(Filtered(MOP.RootNumbers, i->Length(matrix[i])>1));
  for r in [1..Length(matrix)] do
    for op in [1..Length(matrixop)] do
      if not IsBound(matrix[r][op]) then
        matrix[r][op]:=MOP.MultiZero;
      fi;
    od;
  od;

  SubtractSet(CoveredRows, MOP.kPositions(s));

##
## Apart from those we add the position "target" to CoveredRows;
## this may be labelled by "M" or "K". We put "target" into the
## first position of the list to make sure that it will be eliminated
## if this is possible.
##
  RemoveSet(CoveredRows, target);
  CoveredRows:=Concatenation([target], CoveredRows);
  matrix:=matrix{CoveredRows};
#  for x in matrix{[2..Length(matrix)]} do
#    x[1]:=MOP.MultiZero;
#  od;

##
## We build a matrix m whose columns are indexed by the "operators"
## and whose rows are indexed by CoveredRows.
## If op is the Nth operator in the list
## and op acts non-trivially on a root r
## and the string s_N has an "M" or a "K" at position r,
## then we label the entry
## m[r*op][op] by an indeterminate m<r> or k<r>-<N> respectively.
##

##
## Now we form the submatrix consisting of all rows and columns involved
## in the operation. Empty entries are filled with the zero polynomial.
##

  MOP.VerbosePrint(2, "operators: ", record.operators, "\n");

  if MOP.Verbose>=4 then
    Print("operators\n", matrixop, "\n",
          "system of equations:\n");
    PrintArray(matrix);
    Print("\n");
  fi;
##
## We apply the Gaussian elimination procedure to the submatrix.
## For each row we check whether there is a nonzero entry in it.
## If this is the case, then we add the corresponding root to the
## "eliminationlist" which contains all roots that we can eliminate
## by this process. Otherwise we add it to "pollutionlist".
##
  eliminationlist:=[];
  pollutionlist:=[];
  factorlist:=[];

  for r in [1..Length(matrix)] do
    factor:=MOP.ReduceRow(matrix, r, simpleops, ColumnList);

    if factor<>false then
      AddSet(eliminationlist, r);
      UniteSet(factorlist, Set(FactorsInt(AbsInt(factor))));
    else
      AddSet(pollutionlist, r);
    fi;
  od;

  if MOP.Verbose>=4 then
    PrintArray(matrix);
    Print("\n");
  fi;

  newlist:=[];
  repeat
    oldlist:=Copy(newlist);
    for i in Difference(eliminationlist, oldlist) do
      success:=true;
      for j in Difference([1..Length(matrixop)], simpleops) do
        if matrix[i][j] <> MOP.MultiZero then
          if not IsSubset(Union(matrixop{ColumnList{oldlist}}), 
                          matrixop[j]) then
            success:=false;
          fi;
        fi;        
      od;
      if success then
        Add(newlist, i);
      fi;
    od;
  until newlist=oldlist;

  pollutionlist:=Difference([1..Length(matrix)], oldlist);

  return rec(eliminationlist:=CoveredRows{oldlist},
             pollutionlist:=CoveredRows{pollutionlist},
             factorlist:=factorlist);
end;


#############################################################################
##
#F MOP.Elimination(<s>, <target>) . . . . . . . . . . . . . . . . . . .
##
## operates on <s>
## returns <boolean>
##
## The return value indicates whether it is possible to eliminate the
## position <target> in a string <s>.
## If this is possible, the function applies the elimination to the
## string <s>. This operation may change other positions of <s> to "K".
##
MOP.Elimination:=function(s, target)
  local mpos, op, timestamp, new, oplist, record, opnr, root, matrix,
        elimination, index, i, sc, x, newroots, newrecord,
        CurrentSet, NextSet;

  timestamp:=MOP.gettime();
##
## We look for a list of minimum length among the lists of operators
## which are suitable to eliminate the <target> position in <s>.
## Therefore, we start with the empty list and extend it in different
## ways until we find a solution or give up. 
##
  matrix:=List(MOP.RootNumbers, x->[]);
  for i in MOP.mPositions(s) do
    matrix[i][1]:=MOP.Multi(Concatenation("m", String(i)));
  od;
  for i in MOP.kPositions(s) do
    matrix[i][1]:=MOP.Multi(Concatenation("k", String(i)));
  od;

  CurrentSet:=Set([rec(
            matrix:=matrix,
            matrixoperators:=[[]],
## <matrixoperators[i]> is a list which is
## assigned to the i-th column of <matrix>.
            oldroots:=Difference(Set(MOP.kPositions(s)), Set([target])),
            newroots:=Set([target]),
            operators:=[]
           )]);
  NextSet:=Set([]);

  while CurrentSet<>Set([]) do
    for record in CurrentSet do
##
## If more than MOP.MaxEliminationTime seconds are spent on the elimination
## of the <target> position in <s>, then we abort.
##
      if QuoInt((MOP.gettime()-timestamp),1000) > MOP.MaxEliminationTime then
        return false;
      fi;
      if record.operators<>[] then
##
## Check whether it is possible to eliminate the <target> position
## by means of the operators in the list <record.operators>.
##
        elimination:=MOP.Reduce(s, target, record);
        if elimination.eliminationlist<>[] then
##
## If this is the case, we compute the result of the operation.
##
          sc:=MOP.StringCopy(s);
          for i in elimination.eliminationlist do
            sc.k[i]:=false;
            sc.m[i]:=false;
          od;
          for i in elimination.pollutionlist do
            sc.k[i]:=true;
            sc.m[i]:=false;
          od;
##
## In our strategy we permit these new "K"-entries only if their position
## number is greater than <target> or if one of the four abortion criteria
## applies. This strategy is efficient, because in the first case the
## resulting string has been analyzed previously.
##
          if Filtered(elimination.pollutionlist, x->x<target)=[] or
                MOP.FixedJHeight(sc, false) or
                MOP.AlreadyDone(sc, false) or
                MOP.MemberOfInductionList(sc, false) or
                MOP.RankCondition(sc, false) then
            UniteSet(MOP.factorlist, elimination.factorlist);
            if MOP.Verbose>=1 then
              Print("apply ", record.operators, "\n");
              Print("eliminating  ", elimination.eliminationlist, "\n");
              if elimination.pollutionlist<>[] then
                Print("polluting  ", elimination.pollutionlist, "\n");
              fi;
            fi;
##
## Apply the operation to <s> and return.
##
            s.k:=sc.k;
            s.m:=sc.m;
            return true;
          fi;
        fi;
      fi;
##
## Now we are in the case that <record.operators> is not suitable to
## eliminate the <target> position. We try to extend the list of operators
## and save all possible extensions in <NextSet>.
##
## We continue only if the number of operators considered simultaneously is
## suitably small. Otherwise, we do not expect this to be a successful attempt.
##
      if Length(record.operators) <
                         MOP.W.rank+MOP.LowerBound+1 then
##
## We pass through the list of new roots.
##
        for root in record.newroots do
##
## We pass through positions labelled with "m" which are suitable to 
## eliminate the root.
##
          for mpos in Filtered(MOP.RootNumbers, r->
                IsBound(record.matrix[r][1]) and
                Position(String(record.matrix[r][1]), 'm') <> false) do
            opnr:=MOP.operation.fromto[mpos][root];
            if opnr <> 0 then
##
## We extend the operator list by the chosen operator <opnr>.
##
              if not opnr in record.operators then
                newrecord:=Copy(record);
                MOP.ApplyOperator(opnr, target, newrecord);
                Add(newrecord.operators, opnr);
                index:=Length(newrecord.operators)+1;
##
## <newroots> are positions in the string which have been modified
## by the operators and which are supposed to be eliminated at some later
## step. Position numbers which are greater than <target> are not
## added to <newroots> but to the <pollutionlist> instead.
##
                AddSet(newrecord.oldroots, root);
                SubtractSet(newrecord.newroots, newrecord.oldroots);
##
## Save the extended list of operators and the record to the set
## <NextSet> for the next loop.
##
                AddSet(NextSet, Copy(newrecord));
              fi;
            fi;
          od;
        od;
      fi;
    od;
    CurrentSet:=NextSet;
    NextSet:=Set([]);
  od;
  return false;
end;


#############################################################################
##
#F MOP.ApplyOperatorDestroy(<s>, <rootnr>) . . . . . . . . . . . . . . . .
##
## returns a list of strings obtained by application of various
## operators of the form <r>(x) to the string <s>,
## where x is a field element and <r> is the root with number <rootnr>.
##
## The support of any of the resulting strings is not necessarily smaller
## than that of <s>, because here we allow operations which reintroduce
## roots into the support of the resulting string.
##
MOP.ApplyOperatorDestroy:=function(s, opnr)
    local x, rx, y, positions, p, l, st, sc, sl;

##
## We create a list of possible elimination positions.
##
    positions:=List(Filtered(MOP.operation.pairs[opnr],
                             y->(s.k[y[2]] or s.m[y[2]]) and s.m[y[1]]),
                    p->p[2]);
    sc:=MOP.StringCopy(s);
    sl:=[];
##
## We create a temporary string which contains the image of the original
## string under the operation <r>(e) for each e in K. From this temporary
## string we derive new valid strings.
##
## The following table shows the operation of <r>(x) on the positions
## a,b which satisfy a+r=b.
## X is any of 0, M, K.
##
## (entry at a,entry at b) in <s> --> entry at b in <s>*<r>(x)
##------------------------------------------------------------
##          (0,X)                 --> X
##          (K,X)                 --> K
##          (M,0)                 --> M
##          (M,M)                 --> K
##          (M,K)                 --> K
##
    for y in MOP.operation.pairs[opnr] do
        x:=y[1];
        rx:=y[2];
        if sc.k[x] then
            sc.k[rx]:=true;
            sc.m[rx]:=false;
        elif sc.m[x] then
            if sc.m[rx] then
                sc.k[rx]:=true;
                sc.m[rx]:=false;
            elif not sc.k[rx] then
                sc.m[rx]:=true;
                sc.k[rx]:=false;
            fi;
        fi;
    od;
##
## Each position can be eliminated by a suitable field element x.
##
    for p in positions do
        st:=MOP.StringCopy(sc);
        st.k[p]:=false;
        st.m[p]:=false;
        Add(sl, st);
    od;
##
## There exist field elements x such that none of the positions
## is eliminated.
##
    Add(sl, sc);
    return sl;
end;


#############################################################################
##
#F MOP.ApplyOperator(<rootnr>, <target>, <record>). . . . . .
##
## The <string> contains the images of the original string
## under the operation r(x) for each field element x, where r is the root
## with number <rootnr>.
## The <list> consists of the positions whose entries in <s>*r(x) are
## different from those in <s>.
##
MOP.ApplyOperator:=function(rnr, target, record)
  local x, rx, y, i, newop, p;

  for y in MOP.operation.pairs[rnr] do
    x:=y[1];
    rx:=y[2];
    if record.matrix[x]<>[] then
      for i in [1..Length(record.matrixoperators)] do
        if IsBound(record.matrix[x][i]) then
          newop:=Concatenation(record.matrixoperators[i], [rnr]);
          p:=Position(record.matrixoperators, newop);
          if p=false then
            Add(record.matrixoperators, newop);
            p:=Position(record.matrixoperators, newop);
          fi;
          record.matrix[rx][p]:=Copy(record.matrix[x][i]);
          AddSet(record.newroots, rx);
        fi;
      od;
    fi;
  od;
end;


#############################################################################
##
#F MOP.ExtendedAnalysis(<s>) . . . . . . . . . . . . . . . . . . . . .
##
## introduces new positions labelled with "M" in <s> by application
## of some operators and then repeats the usual analysis steps.
##
MOP.ExtendedAnalysis:=function(s)
    local t, opnr, x, diff;

    MOP.Counter.Extended:=MOP.Counter.Extended+1;
    for opnr in Concatenation(
          Filtered(MOP.operation.rootnumbers, n->n>MOP.W.N),
          MOP.RootNumbers) do
        for t in MOP.ApplyOperatorDestroy(s, opnr) do
            MOP.VerbosePrint(1, "Extended\n", s, "\n",
                      "try operator ", opnr, "\n",
                      "try string ", t, "\n");
            if MOP.AnalyzeString(t, false) then
              MOP.VerbosePrint(1, "extended analysis finished successfully\n");
              return;
            fi;
        od;
    od;
    Add(MOP.CannotAnalyzeList, s);

    diff:=MOP.SupportLength(s)-MOP.ComputeRank(s);
    if diff>MOP.MaxModality then
        MOP.MaxModality:=diff;
    fi;

    MOP.VerbosePrint(1, "extended analysis aborted\n");
    return;
end;


#############################################################################
##
#F MOP.Mark(<s>) . . . . . . . . . . . . . . . . . . . .
##
## insert the string <s> into the RedundancyList and also all
## conjugates of <s> under the small Weyl group MOP.WJ
##
MOP.Mark:=function(s)
    local r, t;

#            for t in Orbit(MOP.WJ, s, MOP.OperationOnString) do
    MOP.MarkStringAsDone(s);
    for r in MOP.WJ.generators do
        t:=MOP.OperationOnString(s, r);
        MOP.MarkStringAsDone(t);
    od;
end;

#############################################################################
##
#F MOP.MarkStringAsDone(<s>) . . . . . . . . . . . . . . . . . . . .
##
## inserts a string <s> into the list MOP.RedundancyList,
## provided it is "new" in the sense that it is not a substring
## of any of the existing strings in MOP.RedundancyList. In order to save
## space, we remove all substrings of <s> from MOP.RedundancyList.
##
MOP.MarkStringAsDone:=function(s)
    local e;

    if not MOP.UseRedundancyList then
      return;
    fi;

    for e in MOP.RedundancyList do
        if MOP.IsSubstring(e, s) then
            return;
        fi;
    od;

    MOP.RedundancyList:=Set(Filtered(MOP.RedundancyList, 
                                     e->not MOP.IsSubstring(s, e)));
    AddSet(MOP.RedundancyList, s);
    MOP.VerbosePrint(1, "done:\n", s, "\n");
end;


#############################################################################
##
#F MOP.AnalyzeString(<s>, <normalmode>) .  . . . . . . . . . . . . . . . . .
##
## is the central procedure of this program. Its
## purpose is to reduce the support length of a given string <s> and
## eventually to eliminate it. More precisely, it eliminates roots from
## the support of <s> repeatedly, as long as this is feasible.
## Otherwise a splitting is chosen and the two
## resulting strings are put back onto the stack. As soon as a string
## satisfies one of the deletion criteria it is considered as
## analyzed. If this is not the case and there are no positions left
## in <s> labelled with "K" (i.e. no more splittings or eliminations are
## possible), then we apply the --> MOP.ExtendedAnalysis.
##
## <normalmode> is a <boolean>. In the "normal" mode, MOP.ExtendedAnalysis
## is allowed. In order to avoid loops, this mode is not allowed when
## MOP.AnalyzeString is called from MOP.ExtendedAnalysis.
##
MOP.AnalyzeString:=function(s, normalmode)
    local t, i, kpos, mpos, split, eliminate,
          poslist, target, a;

    MOP.Counter.String:=MOP.Counter.String+1;
    repeat
        mpos:=Reversed(MOP.mPositions(s));
        kpos:=MOP.kPositions(s);
        MOP.VerbosePrint(1, "analyzing\n", s, "\n");

        if MOP.FixedJHeight(s, true) or
           MOP.AlreadyDone(s, true) or
           MOP.MemberOfInductionList(s, true) or
           MOP.RankCondition(s, true) then
            return true;
        fi;

        eliminate:=false;
##
## <poslist> is the list of positions in <s> we want to eliminate.
## If there are positions left in <s> labelled with a "K", then we want
## to eliminate such a position corresponding to a root of minimal height.
## Otherwise, we want to eliminate a position labelled with an "M".
##
        if mpos=[] then
          poslist:=[];
        elif kpos=[] then
          poslist:=mpos;
        else
##          poslist:=Concatenation(mpos, [Minimum(kpos)]);
          poslist:=[Minimum(kpos)];
        fi;
##
## Check whether elimination is possible:
##
        for target in poslist do
          if not eliminate then
            eliminate:=MOP.Elimination(s, target);
          fi;
        od;

        if not eliminate then
##
## If elimination is not possible, the either split or abort.
##
          if kpos=[] then
##
## If splitting is not possible, then perform extended analysis:
##
              if not normalmode then
                  return false;
              fi;
              MOP.ExtendedAnalysis(s);
              return;
          fi;
##
## splitting:
##
          split:=Minimum(kpos);
          s.k[split]:=false;
          MOP.VerbosePrint(1, "splitting    ", split, "\n");

          MOP.Counter.Split:=MOP.Counter.Split+1;
          if normalmode then
            s.m[split]:=true;
            MOP.PushString(s, true);
            s.m[split]:=false;
            MOP.PushString(s, true);
            return;
          else
            t:=MOP.StringCopy(s);
            t.m[split]:=true;
            a:=MOP.AnalyzeString(t, false)
               and MOP.AnalyzeString(s, false);
            if a then
              s.k[split]:=true;
              MOP.Mark(s);
            fi;
            return a;
          fi;
        fi;
    until false;
end;


#############################################################################
##
#F MOP.Initialize(<type>, <rank>, <J>, [rec(...)]) . . . . . . . . .
##
## The first three parameters are mandatory: the letter which indicates
## the type of a Dynkin diagram, the rank of diagram and the subset J
## of {1, ..., rank} which defines the parabolic subgroup P.
##
## The fourth parameter is an optional record. It is used to overwrite
## existing items of the MOP record, e.g. MOP.Verbose. The purpose of
## these global variables is explained below ("optional parameters").
##
MOP.Initialize:=function(arglist)
  local i, Type, rank, J, param;

  Type:=arglist[1];
  rank:=arglist[2];
  J:=arglist[3];

  MOP.Type:=Type;
  MOP.J:=J;
  MOP.SubDiagrams:=[];
  MOP.SubDiagramsDetail:=[];
  MOP.SubsystemOps:=rec(Print:=MOP.PrintSubsystem);
  MOP.TypeOps:=rec(Print:=MOP.PrintType);
##
## optional parameters:

  MOP.Verbose:=0;
## if set > 0, the program shows lots of intermediary results.

  MOP.UseRedundancyList:=true;
## if <true>, the program uses the redundancy list

  MOP.UseInductionList:=true;
## if <true>, the program uses the induction list

  MOP.LowerBound:=0;
## Sometimes a lower bound for the modality is known for theoretical reasons.
## In this case this variable should be set to cut down CPU time and
## to make the rank criterion effective.

  MOP.MaxEliminationTime:=60;
## determines the maximum number of seconds which are spent on one call
## of MOP.Elimination.

  MOP.SaveInterval:=20;
## defines the number of minutes after which the status of the program
## execution is supposed to be saved (--> MOP.WriteStack()). The actual
## saving may occur considerably later, because the program execution must
## first return from all function calls and reach the main loop in
## --> MOP.Modality.

  MOP.RTflag:=0;
  MOP.RTcycles:=0;
## Variables required for Runtime correction. RTflag keeps track of the sign
## of the previous value of Runtime - so overflows can be detected; RTcycles
## holds the number of times the clock has gone through the 2^30 boundary.

  MOP.LastSave:=0;
  MOP.Counter:=rec(Rank:=0, Extended:=0, Redundancy:=0, Split:=0,
                   String:=0, Induction:=0, JHeight:=0);
  MOP.CannotAnalyzeList:=[];
  MOP.factorlist:=[];

##
## RedundancyList is used to keep track of strings that have been
## analyzed so far.
##
  MOP.RedundancyList:=Set([]);

  MOP.W:=CoxeterGroup(Type, rank);
  MOP.RootNumbers:=[1..MOP.W.N];
  MOP.WJ:=ReflectionSubgroup(MOP.W,J);
  MOP.JobName:=Concatenation(MOP.Type, String(rank),
                 Concatenation(List(J, i->
                         Concatenation("_", String(i)))));
  MOP.RootsInRadical:=Filtered(MOP.W.roots, r-> MOP.JHeight(r, J)>=1);
  MOP.PositionsRootsInLevi:=Filtered(MOP.WJ.rootInclusion, x->x<MOP.W.N);
  MOP.PositionsRootsInRadical:=
      Filtered(MOP.RootNumbers, x->not x in MOP.PositionsRootsInLevi);

  if MOP.Verbose>=1 then
    Print("\nType ", Type, rank, ", J = ", J, "\n");
    PrintDynkinDiagram(CartanType(CartanMat(Type, rank)));
    Print("\n", Length(MOP.RootsInRadical), " roots in radical\n\n");
  fi;

##
## optional parameters:
##
  MOP.startstring:=rec();
## define a string to start with. It must be a
##   rec(m:=<blist>, k:=<blist>) ,
## where the boolean list define the positions labelled with "M" or "K", 
## respectively.
## If this parameter is not provided, then it is computed from the
## following:

  MOP.m:=[];
  MOP.k:=MOP.PositionsRootsInRadical;
## define the support of the start string by simple lists
## of numbers of roots which are labelled with "M" or "K", respectively.
## The default values of the lists represent the full Lie algebra of
## the unipotent radical of the parabolic P.

  if Length(arglist) = 4 then
    param:=arglist[4];
    for i in RecFields(param) do
      if IsBound(MOP.(i)) then
        MOP.(i):=param.(i);
        Print("option ", i, ":=", param.(i), "\n");
      else
        Print("unknown option ", i, "\n");
      fi;
    od;
  fi;

  SubtractSet(MOP.m, MOP.k);
  if MOP.startstring=rec() then
    MOP.startstring:=rec(k:=BlistList([1..MOP.W.N], MOP.k),
                         m:=BlistList([1..MOP.W.N], MOP.m));
  fi;

  MOP.startstring.operations:=MOP.StringOps;
  MOP.MaxModality:=MOP.LowerBound;

  MOP.MaxJHeight:=MOP.JHeight(MOP.W.roots[MOP.W.N], J);
  MOP.JHeightStrings:=List([1..MOP.MaxJHeight], h->rec(
                      k:=BlistList(MOP.RootNumbers,
                        Filtered(MOP.RootNumbers, r->
                          MOP.JHeight(MOP.W.roots[r], J)=h)),
                      m:=BlistList(MOP.RootNumbers, [])));

  if MOP.Verbose>=1 then
    for i in MOP.RootNumbers do
        Print(i, ": ", MOP.W.roots[i], "\n");
    od;
  fi;

  MOP.ComputeInductionList();

  MOP.InductionListIndex:=TransposedMat(List(MOP.InductionList, x->
                            BlistList(MOP.RootNumbers, x)));
  MOP.VerbosePrint(1, "Length of induction list = ", Length(MOP.InductionList), "\n");

  MOP.operation:=MOP.ComputeOperation(MOP.W);

##
## The stack is used for keeping track of the strings that still have
## to be analyzed. Whenever a string is put onto the stack, it is put
## there twice. When it is read from the stack the first time, it is
## analyzed (--> MOP.AnalyzeString). When it is encountered the second time,
## then it has been successfully analyzed and thus we may insert it
## into the --> MOP.RedundancyList.
##
    MOP.Stack:=[];
    MOP.StackPointer:=1;
##
## Attempt to read a dump file saved in an earlier run of MOP. 
##
    if MOP.ReadStack() then
        Print("Reading Stack\n");
    else
##
## If this fails, a single string is put onto the stack. 
##
        MOP.PushString(MOP.startstring, true);
        MOP.VerbosePrint(1, "string with ", MOP.SupportLength(MOP.startstring),
                  " entries\n", MOP.startstring, "\n");
    fi;
end;


#############################################################################
##
#F MOP.PrintResult(<r>) . . . . . . . . . . . . . . . . . . . . . .
##
## displays the results of the record <r> computed by
## --> MOP.Modality.
##
MOP.PrintResult:=function(r)
    Print("Modality analysis for type ", r.Type, r.W.rank, 
          ", J = ", r.J, "\n");
    PrintDynkinDiagram(CartanType(CartanMat(r.Type, r.W.rank)));
    Print(
      "--------------------------------------------\n",
      r.Counter.String, " strings analyzed\n",
      r.Counter.Split, " splittings\n",
      "--------------------------------------------\n",
      r.Counter.Induction, " induction list matches\n",
      r.Counter.Redundancy, " already done\n",
      r.Counter.Rank, " occurrences of rank condition\n",
      r.Counter.JHeight, " J-Height criterion invoked\n",
      "--------------------------------------------\n",
      r.Counter.Extended, " extended operations\n",
      "--------------------------------------------\n",
      Length(r.CannotAnalyzeList), " unresolved strings\n",
      "characteristic restrictions: ", r.factorlist, "\n");

    Print("The modality of P is ");
    if r.MaxModality<>r.LowerBound then
      Print("at most ");
    fi;
    Print(r.MaxModality, ".\n",
      "--------------------------------------------\n");
end;


#############################################################################
##
#F Modality(<type>, <rank>, <J>, [rec(...)]) . . . . . . . . .
##
## is the only front end of this package. Its main purpose is to compute
## the modality of the adjoint action of a parabolic on the Lie algebra
## of its unipotent radical.
##
## It has got exactly the same parameters as MOP.Initialize(); see above
## for an explanation.
##
## We allow for instances of type A, D or E only.
##
Modality:=function(arg)
    local e, string, t, r;

    if not Length(arg) in [3,4] then
      Print("Error, usage: Modality(Type, Rank, J [, options])\n");
      return false;
    fi;

    if not arg[1] in ["A", "D", "E"] then
      Print("Error, Type must be A, D or E\n");
      return false;
    fi;

    MOP.Initialize(arg);

    while not MOP.StackEmpty() do
        e:=MOP.PopString();
        string:=e.str;
        if e.stat then
            MOP.PushString(string, false);
            MOP.AnalyzeString(string, true);
#            Print("runtime=", MOP.gettime(),
#                  " lastsave=", MOP.LastSave, "\n");
            if MOP.gettime()>MOP.SaveInterval*60000+MOP.LastSave or
               MOP.Counter.String mod 1000 = 0 then
                MOP.LastSave:=MOP.gettime();
                MOP.WriteStack();
 		if MOP.Verbose >= 1 then
		Print("WriteStack called at ", MOP.LastSave, "\n" );
		fi;
	    else
		if MOP.Verbose >= 1 then
		Print("Writestack NOT called at ", MOP.gettime(), " (Lastsave: ",
		      MOP.LastSave, ")\n" );
		fi;
            fi;
        else
          MOP.Mark(string);
        fi;
    od;

    RemoveSet(MOP.factorlist, 1);

    return rec(
        Counter:=MOP.Counter,
        CannotAnalyzeList:=MOP.CannotAnalyzeList,
        factorlist:=MOP.factorlist,
        Type:=MOP.Type,
        J:=MOP.J,
        LowerBound:=MOP.LowerBound,
        W:=MOP.W,
        WJ:=MOP.WJ,
        MaxModality:=MOP.MaxModality,
        SubDiagrams:=MOP.SubDiagrams,
        SubDiagramsDetail:=MOP.SubDiagramsDetail,
        operations:=rec(Print:=MOP.PrintResult)
    );
end;


#
# multivariate polynomials
#

MOP.MultiOps:=rec();


MOP.MultiZero:=rec(
  operations:=MOP.MultiOps,
  polynomial:=Set([]),
  isMulti:=true);


MOP.MultiOne:=rec(
  operations:=MOP.MultiOps,
  polynomial:=Set([rec(Bscalar:=1, Afactors:=Set([]))]),
  isMulti:=true);


MOP.MultiScalar:=function(z)
  local p;

  if z=0 then
    return Copy(MOP.MultiZero);
  fi;
  p:=Copy(MOP.MultiOne);
  p.polynomial[1].Bscalar:=z;
  return p;
end;


MOP.MultiOps.name:="multi";


MOP.IsMulti := function( obj )
    return IsRec(obj) and IsBound(obj.isMulti) and obj.isMulti;
end;


MOP.MultiOps.Print:=function(p)
  Print(String(p));
end;


MOP.MultiOps.String:=function(p)
  local i, j, s;

  if p=MOP.MultiZero then
    return "0";
  fi;

  s:="";
  for i in p.polynomial do
    if i.Bscalar>0 then
      Append(s, "+");
    fi;
    Append(s, String(i.Bscalar));
    for j in i.Afactors do
      Append(s, "(");
      Append(s, j.Avar);
      Append(s, ")^");
      Append(s, String(j.Bexp));
      Append(s, " ");
    od;
  od;
  return s;
end;


MOP.MultiOps.\*:=function(a, b)
  local ma, mb, factors, afactor, bfactor, found, result;

  if IsInt(a) then
    return MOP.MultiScalar(a)*b;
  fi;

  if IsInt(b) then
    return a*MOP.MultiScalar(b);
  fi;

  if not MOP.IsMulti(a) or not MOP.IsMulti(b) then
    return false;
  fi;

  if a=MOP.MultiZero or b=MOP.MultiZero then
    return Copy(MOP.MultiZero);
  fi;

  if IsMonomial(a) and IsMonomial(b) then
    ma:=a.polynomial[1];
    mb:=b.polynomial[1];
     
    factors:=Copy(ma.Afactors);

    for bfactor in mb.Afactors do
      found:=false;
      for afactor in factors do
        if afactor.Avar=bfactor.Avar then
          afactor.Bexp:=afactor.Bexp+bfactor.Bexp;
          found:=true;
        fi;
      od;
      if not found then
        Add(factors, Copy(bfactor));
      fi;
    od;

    result:=rec(operations:=MOP.MultiOps, isMulti:=true, polynomial:=Set([
              rec(Bscalar:=ma.Bscalar*mb.Bscalar, Afactors:=Set(factors))]));
  else
    result:=Sum(a.polynomial, ma->Sum(b.polynomial, mb->
       rec(operations:=MOP.MultiOps, isMulti:=true, polynomial:=Set([ma])) *
       rec(operations:=MOP.MultiOps, isMulti:=true, polynomial:=Set([mb]))));
  fi;
  return result;
end;


MOP.MultiOps.\-:=function(a, b)
  return a+(-1)*b;
end;



MOP.MultiOps.\+:=function(a, b)
  local ma, mb, p, found, new;

  p:=Copy(a);
  Unbind(p.isMonomial);

  for mb in b.polynomial do
    found:=false;
    for ma in p.polynomial do
      if mb.Afactors=ma.Afactors then
        found:=true;
        ma.Bscalar:=ma.Bscalar+mb.Bscalar;
      fi;
    od;
    if not found then
      Add(p.polynomial, Copy(mb));
    fi;
  od;

  p.polynomial:=Set(Filtered(p.polynomial, x->x.Bscalar<>0));
  return p;
end;


MOP.MultiOps.IsMonomial:=function(p)
  return MOP.IsMulti(p) and Size(p.polynomial)=1;
end;


MOP.Multi:=function(x)
  local p;

  p:=rec();
  p.operations:=MOP.MultiOps;
  p.polynomial:=Set([rec(Bscalar:=1, 
                         Afactors:=Set([rec(Avar:=String(x), Bexp:=1)]))]);
  p.isMulti:=true;
  return p;
end;

#############################################################################
##
## Function MOP.getime
##
## M.Finney January 2003
## to act as a replacement for GAP Runtime() so as to avoid the overflow
## problems inherent in the Runtime function (small integer return, 
## overflow at 2^28, negative at 2^29, etc.
##
############################################################################# 
MOP.gettime:=function()
  local moptime;
  ## normalise moptime by adding 0
  moptime:=Runtime() + 0;
  if SignInt(moptime) = -1 and MOP.RTflag <> -1 then
	MOP.RTcycles:= MOP.RTcycles + 1;
	fi;
  MOP.RTflag:=SignInt(moptime);

  return (MOP.RTcycles * 2^30) + moptime;
end;



